# -*- coding: utf-8 -*-
# author: NhanDD3 <hp.duongducnhan@gmail.com>
import logging
import os
import json
import requests
from functools import wraps
from typing import Dict, List
from typing import Callable
from pathlib import Path
from logging.handlers import TimedRotatingFileHandler
from pythonjsonlogger import jsonlogger


def json_translate_obj(obj):
    # for example, serialize a custom object
    # if isinstance(obj, MyClass):
    #     return {"special": obj.special}
    return {'obj': str(obj)}


def setup_time_rotation_logger(
    name, 
    level, 
    log_file_dir=None, 
    format_json=False, 
    json_translator=json_translate_obj
):
    if not log_file_dir:
        log_file_dir = os.path.join(Path.home(), "cmk-tools", "logs")

    if not os.path.isdir(log_file_dir):
        os.makedirs(log_file_dir, exist_ok=True)
    
    log_file_path = os.path.join(log_file_dir, f"{name}.log")
    # print("log files will be saved at:", log_file_path)

    logger = logging.getLogger(f'cmk-tools.{name}')
    logger.setLevel(logging.DEBUG)

    handler = TimedRotatingFileHandler(
        log_file_path, when="midnight", interval=1, backupCount=7
    )

    if format_json:
        formatter = jsonlogger.JsonFormatter(
            json_encoder=json.JSONEncoder,
            json_default=json_translator,
        )
    else:
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    handler.setFormatter(formatter)
    handler.setLevel(level)
    logger.addHandler(handler)
    return logger


def setup_log(name: str, level: int = logging.INFO):
    home_path = Path.home()
    log_dir = os.path.join(home_path, "fmon_active_check_logs")

    active_check_log_dir_name = name
    if active_check_log_dir_name.endswith([".log", ".txt"]):
        active_check_log_dir_name = active_check_log_dir_name.split(".")[0]
    log_dir = os.path.join(log_dir, active_check_log_dir_name)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    logger = setup_time_rotation_logger(name, level, log_file_dir=log_dir, format_json=True)
    logger.debug = restruct_log_method(logger.debug)
    logger.info = restruct_log_method(logger.info)
    logger.warn = restruct_log_method(logger.warn)
    logger.error = restruct_log_method(logger.error)
    logger.critical = restruct_log_method(logger.critical)
    logger.exception = restruct_log_method(logger.exception, True)
    logger.warning = restruct_log_method(logger.warning)
    return logger

def restruct_log_method(func, exc_info: bool = None):
    @wraps(func)
    def structured_method(
        message,
        exc_info: bool = exc_info,
        extra: Dict = None,
        stack_info=False,
        stacklevel=1,
        **kwargs
    ):
        if not extra:
            extra = {}

        for k, v in kwargs.items():
            if k == 'api_resp' and isinstance(v, requests.Response):
                extra.update({
                    'api_request_url': v.request.url,
                    'api_request_method': v.request.method,
                    'api_request_headers': dict(v.request.headers),
                    'api_request_body': v.request.body.decode('utf-8'),
                    'api_status_code': v.status_code,
                    'api_response_text': v.text
                })
        try:
            return func(message, exc_info=exc_info, extra=extra, stack_info=stack_info, stacklevel=stacklevel)
        except Exception as e:
            new_extra = {'exc_extra': str(extra)}
            return func(message, exc_info=exc_info, extra=new_extra, stack_info=stack_info, stacklevel=stacklevel)
    return structured_method



# -*- coding: utf-8 -*-
# author: NhanDD3 <hp.duongducnhan@gmail.com>

import logging
import time
import uuid
import redis
from .utils import get_func_name


logger = logging.getLogger("cmkredistools")


class RedisSemaphore:
    def __init__(self, redis_url: str, name, limit=-1, timeout=10):
        self.name = name  # Name of the semaphore key in Redis
        self.limit = limit  # Semaphore limit
        self.df_limit = 100  # Default limit
        self.redis_url = redis_url
        self.redis: redis.Redis = redis.from_url(
            redis_url, socket_timeout=5, decode_responses=True
        )  # Redis connection object
        self.timeout = timeout  # Timeout for acquiring the semaphore
        self.ttl = 2 * timeout + 1  # TTL for each semaphore lock (deadlock prevention)
        if self.is_redis_connected:
            logger.info(f"Connected to Redis")
        else:
            logger.error(f"Failed to connect to Redis at {redis_url}")

        if self.limit < 0:
            self.get_limit()
        else:
            self.set_limit()

        logger.info(
            f"Semaphore {self.name} initialized with limit: {self.limit} with key {self._limit_key_name()}, timeout: {self.timeout}, TTL: {self.ttl}."
        )

    def set_limit(self, limit=None) -> int:
        if not limit:
            limit = self.limit
        # set limit to redis
        limit_key = self._limit_key_name()
        return self.redis.set(limit_key, limit)

    def get_limit(self) -> int:
        # get limit from redis
        limit_key = self._limit_key_name()
        str_limit = self.redis.get(limit_key)
        if not str_limit:
            self.set_default_limit()
            self.limit = self.df_limit
        else:
            try:
                self.limit = int(str_limit)
            except ValueError:
                logger.error(
                    f"Invalid limit value: {str_limit}, set default limit = {self.df_limit}."
                )
                self.limit = self.df_limit  # Set default limit
        return self.limit

    def set_default_limit(self):
        limit_key = self._limit_key_name()
        return self.redis.set(limit_key, self.df_limit)

    def _limit_key_name(self):
        return f"{self.name}:limit"

    def _lock_name(self, lock_id):
        return f"{self.name}:{lock_id}"

    @property
    def is_redis_connected(self):
        if not self.redis:
            raise ValueError("No Redis connection found.")
        return self.redis.ping()

    def count(self):
        """
        Get the current count of the semaphore.
        """
        current_value = self.redis.get(self.name)
        return int(current_value) if current_value else 0

    def clear(self):
        """
        Clear the semaphore count and all lock IDs.
        """
        pipeline = self.redis.pipeline(True)
        pipeline.delete(self.name)
        pipeline.delete(self._lock_name("*"))
        pipeline.delete(self._limit_key_name())
        pipeline.execute()

    def acquire(self, timeout: float = None):
        """
        Acquire a semaphore slot with a unique identifier and set TTL to avoid deadlock.
        """
        if not timeout:
            timeout = self.timeout

        # Generate a unique ID for the semaphore slot
        lock_id = f'{uuid.uuid4()}___{int(time.time())}'
        start_time = time.time()
        while True:
            current_value = self.redis.get(self.name)
            current_value = int(current_value) if current_value else 0

            if current_value < self.limit:
                pipeline = self.redis.pipeline(True)
                pipeline.incr(self.name)  # Increment the semaphore count
                pipeline.expire(self.name, self.ttl)  # Set a TTL for the semaphore count
                # Set a unique lock ID with an expiration (TTL)
                pipeline.set(self._lock_name(lock_id), '1', ex=self.ttl, nx=True)
                pipeline.keys(self._lock_name("*"))  # Get all lock IDs

                # result = [semaphore_count, set_log_id_res, lock_id_keys,]
                result = pipeline.execute()

                if result[0] <= self.limit:  # Check if we acquired the semaphore successfully
                    return lock_id  # Return the unique lock ID
                else:
                    keys_to_delete = []
                    decr_count = 1
                    if result[2]:
                        for key in result[2]:
                            key_time = key.split("___")[-1]
                            if time.time() - float(key_time) > self.ttl:
                                keys_to_delete.append(key)
                                decr_count += 1
                    logger.debug(
                        f'found {len(keys_to_delete)} keys to delete', 
                        extra={
                            'keys_to_delete': keys_to_delete,
                            'decr_count': decr_count,
                        }
                    )
                    # Revert the count increment if the semaphore limit was exceeded
                    if decr_count > 1:
                        pipeline = self.redis.pipeline(True)
                        pipeline.decr(self.name, decr_count)
                        for key in keys_to_delete:
                            pipeline.delete(key)
                        pipeline.execute()
                    else:
                        self.redis.decr(self.name)
            else:
                if self.timeout > 0 and time.time() - start_time >= self.timeout:
                    logger.warning(f'Timeout lock {lock_id} acquiring the semaphore after {timeout} seconds.')
                    return None  # Timeout occurred
                time.sleep(0.05)  # Sleep briefly to avoid busy-waiting

    def release(self, lock_id):
        """
        Release the semaphore only if the lock_id matches to ensure proper release.
        """
        lock_name = self._lock_name(lock_id)
        if self.redis.exists(lock_name):
            pipeline = self.redis.pipeline(True)
            pipeline.decr(self.name)  # Decrement the semaphore count
            pipeline.delete(lock_name)  # Remove the lock ID key
            pipeline.execute()
            logger.info(f"Lock ID {lock_id} released.")
        else:
            logger.warning(f"Lock ID {lock_id} not found or already released.")


def run_with_semaphore(
    semaphore: RedisSemaphore,
    func,
    func_args=(),
    func_kwargs={},
    execute_when_timeout=True,
):
    """
    Run a function with a semaphore acquired and released automatically.
    """
    if not isinstance(semaphore, RedisSemaphore):
        raise ValueError(
            "The 'semaphore' argument must be an instance of RedisSemaphore."
        )

    lock_id = semaphore.acquire()
    if lock_id:
        try:
            logger.info(
                f"Acquired the semaphore for Function {get_func_name(func)} with lock ID: {lock_id}"
            )
            return func(*func_args, **func_kwargs)
        finally:
            semaphore.release(lock_id)
    else:
        if execute_when_timeout:
            logger.warning(
                f"Failed to acquire the semaphore (timeout). Function {get_func_name(func)} executed anyway"
            )
            return func(*func_args, **func_kwargs)
        else:
            logger.warning(
                f"Failed to acquire the semaphore (timeout). Function {get_func_name(func)} NOT executed."
            )


def run_with_semaphore_decorator(semaphore: RedisSemaphore):
    """
    Decorator to run a function with a semaphore acquired and released automatically.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            return run_with_semaphore(semaphore, func, args, kwargs)

        return wrapper

    return decorator


# Usage example
if __name__ == "__main__":
    redis_conn = redis.StrictRedis(host="localhost", port=6379, db=0)
    semaphore = RedisSemaphore(
        "redis://localhost:6379/0", name="my_semaphore", limit=10, timeout=10
    )

    lock_id = semaphore.acquire()
    if lock_id:
        try:
            print(f"Acquired the semaphore with lock ID: {lock_id}, performing task...")
            # Do some work here
            time.sleep(2)  # Simulate a task
        finally:
            semaphore.release(lock_id)
            print(f"Released the semaphore with lock ID: {lock_id}.")
    else:
        print("Failed to acquire the semaphore (timeout).")



import sys
import re
import os
import logging
from typing import Callable
from .utils import replace_domain
from .status import OK, WARNING, CRITICAL, UNKNOWN
from .status import terminate_check as _terminate_check
import requests
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


KEY_ACTIVE_CHECK_ID = 'ACTIVE_CHECK_ID'
# SCC_USE_TEST_SERVER = 'http://testserver/'
SCC_USE_TEST_SERVER = os.getenv("SCC_USE_TEST_SERVER")
if SCC_USE_TEST_SERVER and not 'http' in SCC_USE_TEST_SERVER:
    raise ValueError("SCC_USE_TEST_SERVER must be a valid URL")

def is_test_device(name: str) -> bool:
    try:
        if name and name.lower().startswith('SCC_TEST_SVR_'):
            search_res = re.search(r'SCC_TEST_SVR_([\w\d\.\:]*)_', name)
            if search_res:
                domain_or_ipport = search_res.group(1)
                if domain_or_ipport:
                    if not domain_or_ipport.startswith('http'):
                        # because of local using, we need to add https
                        domain_or_ipport = f'http://{domain_or_ipport}'
                    os.environ['SCC_USE_TEST_SERVER'] = domain_or_ipport
                    return True
    except Exception as e:
        return False

def make_request(
    plugin_id: str,
    method: str, 
    url, 
    params: dict = None,
    data: dict =None,
    json = None,
    headers: dict = {}, 
    verify=False, 
    terminate_check: Callable =_terminate_check,
    proxies: dict = None,
    logger_name: str = "cmk-tools",
    **kwargs
): 
    logger = logging.getLogger(os.environ.get('KEY_ACTIVE_CHECK_ID', logger_name))
    if not callable(terminate_check):
        raise ValueError("terminate_check must be a function")

    method_upper = method.upper()
    if method_upper not in ['GET', 'POST', 'PUT', 'DELETE']:
        return terminate_check(UNKNOWN, f"request API method >>{method}<< not supported")
    
    response = None
    try:
        if SCC_USE_TEST_SERVER:
            new_url = f'{SCC_USE_TEST_SERVER}/cmk-plugin/api/testcase/response/'
            headers = {
                **headers, 
                'x-plugin-id': plugin_id,
                'x-original-url': url,
            }
            response = requests.request('POST', new_url, headers=headers, data=data, verify=verify, proxies=proxies, json=json)
        else:
            if method_upper in ('GET', 'DELETE'):
                response = requests.request(method, url, headers=headers, params=params, verify=verify, proxies=proxies) 
            else:
                response = requests.request(method, url, headers=headers, data=data, verify=verify, json=json, proxies=proxies)
        response.raise_for_status()
        logger.debug(f"Request to {url} with method {method} successfully", api_resp=response)
    except requests.exceptions.ConnectTimeout:
        logger.warning(f"ConnectTimeout when connect to {url}", api_resp=response)
        terminate_check(UNKNOWN, f"Timeout when connect to {url}")
    except requests.exceptions.ReadTimeout:
        logger.warning(f"ReadTimeout when connect to {url}", api_resp=response)
        terminate_check(UNKNOWN, f"Timeout when read data from {url}")
    except Exception as e:
        logger.exception(f"Get exception {e} when connect to {url}", api_resp=response)
        terminate_check(UNKNOWN, f"Error when request to {url} with error: {e}")
    finally:
        return response


# -*- coding: utf-8 -*-
# author: NhanDD3 <hp.duongducnhan@gmail.com>

import re


def get_func_name(func):
    try:
        return f"<{func.__name__}>"
    except AttributeError:
        return f"<{str(func)}>"

def replace_domain(url, domain):
    new = re.sub(r"http[s]*?://[^/ | ^?]+", domain, url)
    return new


#!/usr/bin/env python3
# Template Author: NhanDD <hp.duongducnhan@gmail.com>

"""
Simple Packer and (later) unpacker for MKP Files
MKP is the Package format for Check_MK
"""
import os
import json
import tarfile


def pack_mkp(
    package_name: str,
    checks_file: str,
    lib_file: str,
    wato_file: str,
    checks_file_path: str = None,
    lib_file_path: str = None,
    wato_file_path: str = None,
    base_path: str = os.getcwd()
):
    """
    Pack a MKP File from the given files
    - info
    - info.json
    - active_check 
    - check_invoker
    - wato
    -> version get from 'info' file
    """
    if not checks_file_path:
        checks_file_path = os.path.join(base_path, 'check_invoker', checks_file)
    if not lib_file_path:
        lib_file_path = os.path.join(base_path, 'active_check', lib_file)
    if not wato_file_path:
        wato_file_path = os.path.join(base_path, 'wato', wato_file)

    with open('info', 'r') as f:
        info = f.read()
        info_dict = eval(info)
    if not isinstance(info_dict, dict):
        raise ValueError("Info File is invalid")
    
    info_dict['files'] = {}

    with open('temp_info', 'w') as info_file:
        info_file.write(str(info_dict))
    
    with open('temp_info.json', 'w') as info_json_file:        
        info_json_file.write(json.dumps(info_dict, indent=4))
        
    filename = f"{package_name}-{info_dict.get('version', '1.0.0')}"
    if os.path.isfile(filename + '.mkp'):
        os.remove(filename + '.mkp')
        
    with tarfile.open(filename + '.mkp', "w:gz") as mkp_tar:
        with tarfile.open('checks.tar', 'w:gz') as checks_tar:
            checks_tar.add(checks_file_path, arcname=f"checks/{checks_file}")
            
        with tarfile.open('lib.tar', 'w:gz') as lib_tar:
            lib_tar.add(lib_file_path, arcname=f"agents/plugins/{lib_file}")
            
        with tarfile.open('web.tar', 'w:gz') as web_tar:
            web_tar.add(wato_file_path, arcname=f"web/plugins/wato/{wato_file}")
        
        for tar_file in ['checks.tar', 'lib.tar', 'web.tar']:
            mkp_tar.add(tar_file, arcname=tar_file)
            os.remove(tar_file)
        
        # Add check file to the 'checks' directory
        mkp_tar.add(checks_file_path, arcname=f"checks/{checks_file}")
        
        # Add active check file to the 'agents/plugins' directory
        mkp_tar.add(lib_file_path, arcname=f"agents/plugins/{lib_file}")
        
        # Add WATO plugin file to the 'web/plugins/wato' directory
        mkp_tar.add(wato_file_path, arcname=f"web/plugins/wato/{wato_file}")
        
        # Add info files
        mkp_tar.add('temp_info', arcname='info')
        mkp_tar.add('temp_info.json', arcname='info.json')
        
    os.remove('temp_info')
    os.remove('temp_info.json')


if __name__ == "__main__":
    pack_mkp()


# -*- coding: utf-8 -*-
# author: NhanDD3 <hp.duongducnhan@gmail.com>


__version__ = '1.1.3'


from .logger import setup_time_rotation_logger, setup_log
from .redis_semaphore import (
    RedisSemaphore,
    run_with_semaphore,
    run_with_semaphore_decorator,
)
from .rest_request import make_request, is_test_device
from .status import terminate_check, OK, WARN, WARNING, CRITICAL, ERROR, UNKNOWN
from .mkp_packer import pack_mkp


__all__ = [
    "RedisSemaphore",

    "setup_time_rotation_logger",
    "setup_log",

    "run_with_semaphore",
    "run_with_semaphore_decorator",

    "make_request",
    'is_test_device',

    "terminate_check",
    "OK",
    "WARN",
    "WARNING",
    "CRITICAL",
    "ERROR",
    "UNKNOWN",
    
    'pack_mkp',
]



import sys 


DEBUG = "DEBUG"
OK = "OK"
WARN = "WARN"
WARNING = "WARNING"
CRITICAL = "CRITICAL"
ERROR = "ERROR"
UNKNOWN = "UNKNOWN"



def terminate_check(status, msg):
    # -----------------------------------------   
    #   DO NOT MODIFY THIS FUNCTION
    # -----------------------------------------    
    print(f"{status} - {msg}")
    if status == OK:
        sys.exit(0)
    elif status == WARNING:
        sys.exit(1)
    elif status == CRITICAL:
        sys.exit(2)
    else:
        sys.exit(3)
